import json
import logging
from datetime import datetime
from importlib.resources import files
from pathlib import Path
import random

from dotenv import load_dotenv
from ollama import chat

from opensyndrome.filtering import get_schema_filepath
from opensyndrome.schema import OpenSyndromeCaseDefinitionSchema

load_dotenv()
logger = logging.getLogger(__name__)
DEFAULT_MODEL = "mistral"


def load_examples(examples_dir: Path, random_k=None):
    json_definitions = {}
    for raw_json in examples_dir.glob("**/*"):
        if not raw_json.name.endswith(".json"):
            continue
        if raw_json.read_text() != "":
            content = json.loads(raw_json.read_text())
            if content:
                json_definitions[raw_json.stem] = content

    definitions = list(json_definitions.values())
    if random_k:
        definitions = random.sample(definitions, random_k)
    examples = "\n".join(
        f"- {json.dumps(_definition)}" for _definition in definitions if _definition
    )
    return examples


PROMPT_TO_MACHINE_READABLE_FORMAT = """
You are an expert in creating standardized case definition JSONs for medical syndromes.
Generate a JSON that strictly follows this JSON schema, using the provided example documents as reference.

Strict Rules:
- ONLY use symptoms explicitly mentioned in the input text
- ONLY use criteria explicitly mentioned in the schema
- Generate JSON matching the provided schema exactly
- Do not add any information not in the source text
- Use logical operators to capture text's precise meaning
- If text is ambiguous, minimize assumptions

Example documents to reference:
{examples}

Input: {human_readable_definition}

Expected Output Format:
- Use {language} language
- JSON matching provided schema only
- Criteria reflecting ONLY input text
- No additional professional judgment or external information
"""
PROMPT_TO_HUMAN_READABLE_FORMAT = """
You are an public health expert in creating standardized case definition.
Generate a human-readable definition from the provided JSON using only
the information provided there.

Expected output format:
- Text in narrative form
- Clear, concise, and easy to understand text
- No additional professional judgment or external information
- Write in {language}

{machine_readable_definition}
"""


def _add_first_level_required_fields(schema: dict, definition: dict):
    """Add mandatory fields and empty values as placeholders."""
    default_values = {
        "string": "",
        "array": [],
        "object": {},
        "integer": 0,
    }
    missing_fields = set(schema["required"]) - set(definition.keys())
    for field in missing_fields:
        definition[field] = default_values.get(schema["properties"][field]["type"])
    return definition


def _fill_automatic_fields(
    machine_readable_definition: dict, human_readable_definition: str
):
    machine_readable_definition["human_readable_definition"] = human_readable_definition
    machine_readable_definition["published_in"] = (
        "https://opensyndrome.org/definitions/<replace-url>"
    )
    machine_readable_definition["published_at"] = str(
        datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    )
    machine_readable_definition["published_by"] = []
    machine_readable_definition["status"] = "draft"
    machine_readable_definition["open_syndrome_version"] = (
        "1.0.0"  # TODO get this version from definition repo
    )
    machine_readable_definition["references"] = [
        {"citation": "", "url": ""}
    ]  # to be filled by the user
    schema = json.loads(get_schema_filepath().read_text())
    machine_readable_definition = _add_first_level_required_fields(
        schema, machine_readable_definition
    )
    return machine_readable_definition


def _drop_regex_pattern(node: dict):
    """Recursively drop 'pattern' keys from the schema since it is not supported.

    Issue: https://github.com/ollama/ollama-python/issues/541"""
    original_node = node.copy()
    dropped = node.pop("pattern", None)
    if dropped is not None:
        logger.warning(f"Dropped 'pattern' from {original_node}")
    for value in node.values():
        if isinstance(value, dict):
            _drop_regex_pattern(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    _drop_regex_pattern(item)


def generate_machine_readable_format(
    human_readable_definition, model=DEFAULT_MODEL, language="American English"
):
    if not human_readable_definition:
        raise ValueError("Human-readable definition cannot be empty.")

    examples_dir = files("opensyndrome").joinpath("examples")
    examples = load_examples(examples_dir, 3)
    formatted_prompt = PROMPT_TO_MACHINE_READABLE_FORMAT.format(
        examples=examples,
        human_readable_definition=human_readable_definition,
        language=language,
    )

    json_schema = OpenSyndromeCaseDefinitionSchema.model_json_schema()
    _drop_regex_pattern(json_schema)
    response = chat(
        messages=[{"role": "user", "content": formatted_prompt}],
        model=model,
        format=json_schema,
        options={"temperature": 0},
        stream=False,
    )

    machine_readable_definition = json.loads(response.message.content)
    if isinstance(machine_readable_definition, list):
        if len(machine_readable_definition) > 1:
            logger.warning("More than one definition generated...")
        machine_readable_definition = machine_readable_definition[0]

    return _fill_automatic_fields(
        machine_readable_definition, human_readable_definition
    )


def _exclude_metadata_fields(definition: dict):
    """Exclude metadata fields from the definition."""
    definition_fields = [
        "inclusion_criteria",
        "exclusion_criteria",
        "target_public_health_threats",
    ]
    return {key: value for key, value in definition.items() if key in definition_fields}


def generate_human_readable_format(
    machine_readable_definition, model=DEFAULT_MODEL, language="American English"
):
    if not machine_readable_definition:
        raise ValueError("Machine-readable definition cannot be empty.")

    formatted_prompt = PROMPT_TO_HUMAN_READABLE_FORMAT.format(
        language=language,
        machine_readable_definition=_exclude_metadata_fields(
            machine_readable_definition
        ),
    )
    response = chat(
        messages=[{"role": "user", "content": formatted_prompt}],
        model=model,
        options={"temperature": 0},
        stream=False,
    )
    human_readable_definition = response.message.content
    return human_readable_definition



import json
from functools import wraps
from pathlib import Path

from pygments import highlight, lexers, formatters
import jsonschema
import click
import ollama
from ollama._types import ResponseError

from opensyndrome.converters import (
    generate_machine_readable_format,
    generate_human_readable_format,
)
from opensyndrome.filtering import get_schema_filepath, get_definition_dir
from opensyndrome.validators import validate_machine_readable_format


@click.group()
def cli():
    pass


def validate_machine_readable_format_with_style(json_or_file, schema_file=None):
    try:
        validate_machine_readable_format(json_or_file, schema_file)
        click.echo(click.style("✅ Validation successful!", fg="green"))
    except (json.JSONDecodeError, json.decoder.JSONDecodeError) as e:
        click.echo(click.style(f"❌ Invalid JSON: {e}", fg="red"), err=True)
    except jsonschema.exceptions.ValidationError as e:
        click.echo(click.style(f"❌ Validation error: {e}", fg="red"), err=True)
    except Exception as e:
        click.echo(
            click.style(f"❌ An unexpected error occurred: {e}", fg="red"), err=True
        )


@cli.command("validate")
@click.argument("json_file", type=click.Path(exists=True))
@click.option("--schema-file", type=click.Path(exists=True))
def validate_json(json_file, schema_file):
    """
    Validate a JSON file against a JSON Schema.

    JSON_FILE: Path to the JSON file to validate.
    SCHEMA_FILE: Path to the JSON Schema file. If not passed, it will use
    the downloaded schema from GitHub repo.
    """
    validate_machine_readable_format_with_style(json_file, schema_file)


def color_json(json_definition: dict):
    formatted_json = json.dumps(json_definition, indent=4)
    return highlight(formatted_json, lexers.JsonLexer(), formatters.TerminalFormatter())


def is_ollama_available():
    try:
        ollama.list()
        return True
    except (ConnectionError, ResponseError):
        return False


def check_ollama(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not is_ollama_available():
            click.echo(
                click.style("Ollama service is missing or unavailable.", fg="red"),
                err=True,
            )
            return
        else:
            return func(*args, **kwargs)

    return wrapper


@cli.command("convert")
@click.option(
    "--validate", is_flag=True, help="Validate the JSON file against the schema."
)
@click.option(
    "--model",
    type=str,
    help="Model used to generate the JSON file.",
    default="mistral",
)
@click.option(
    "--language",
    type=str,
    help="Language used to generate the machine-readable definition.",
    default="American English",
)
@click.option(
    "--edit",
    is_flag=True,
    help="Open editor after generation.",
)
@click.option(
    "-hr",
    "--human-readable-definition",
    type=str,
    help="Human-readable definition. If not provided, an editor will open to input the definition.",
)
@check_ollama
def convert_to_json(validate, model, language, edit, human_readable_definition):
    """
    Convert human-readable definition (TEXT) to the machine-readable format (JSON).

    If the --validate flag is passed, the JSON file will be validated against the schema.
    """
    if not human_readable_definition:
        human_readable_definition = click.edit(extension=".json")
    machine_readable_definition = generate_machine_readable_format(
        human_readable_definition, model, language
    )

    if edit:
        machine_readable_definition_edited = click.edit(
            text=json.dumps(machine_readable_definition, indent=4), extension=".json"
        )
        if machine_readable_definition_edited:
            machine_readable_definition = json.loads(machine_readable_definition_edited)

    click.echo(color_json(machine_readable_definition))

    if validate:
        validate_machine_readable_format_with_style(machine_readable_definition)


@cli.command("humanize")
@click.argument("json_file", type=click.Path(exists=True))
@click.option(
    "--model",
    type=str,
    help="Model used to generate the JSON file.",
    default="mistral",
)
@click.option(
    "--language",
    type=str,
    help="Language used to generate the human-readable definition.",
    default="American English",
)
@check_ollama
def convert_to_text(json_file, model, language):
    """Convert a machine-readable format (JSON) to a human-readable format (TEXT)."""
    machine_readable_definition = json.loads(Path(json_file).read_text())
    text = generate_human_readable_format(machine_readable_definition, model, language)
    click.echo(click.style(text, fg="green"))


@cli.command("download", help="Download an entity from OSI.")
@click.argument("entity")
def download_entity(entity):
    match entity:
        case "schema":
            result = get_schema_filepath()
        case "definitions":
            result = get_definition_dir()
        case _:
            result = None

    if not result:
        click.echo(
            click.style(
                f"Invalid entity: {entity}. Expected: `schema` or `definitions`.",
                fg="red",
            )
        )
    else:
        click.echo(click.style(f"{entity} available at: {result}", fg="green"))


def main():
    cli()


if __name__ == "__main__":
    main()



from unittest.mock import Mock

from opensyndrome.cli import is_ollama_available
import pytest
from ollama._types import ResponseError


class TestIsOllamaAvailable:
    @pytest.fixture
    def mock_ollama(self, mocker):
        _mock_ollama = mocker.patch("opensyndrome.cli.ollama")
        list_response = Mock(models=[Mock(model="mistral")])

        def make_mock(return_value=None, side_effect=None):
            if side_effect:
                _mock_ollama.list.side_effect = side_effect
            else:
                _mock_ollama.list.return_value = return_value or list_response
            return _mock_ollama

        return make_mock

    def test_is_ollama_available(self, mock_ollama):
        assert is_ollama_available() is True

    def test_show_error_when_the_model_is_not_available(self, mock_ollama):
        error_message = "model 'gemma3' not found (status code: 404)"
        mock_ollama(side_effect=ResponseError(error=error_message))
        assert is_ollama_available() is False

    def test_show_error_when_cannot_connect_to_ollama(self, mock_ollama):
        mock_ollama(side_effect=ConnectionError())
        assert is_ollama_available() is False



import json
from functools import reduce
from pathlib import Path
import polars as pl
import requests


OPEN_SYNDROME_VERSION = "v1"
OPEN_SYNDROME_DIR = Path.home() / ".open_syndrome" / OPEN_SYNDROME_VERSION
OPEN_SYNDROME_DIR.mkdir(parents=True, exist_ok=True)
SCHEMA_DIR = OPEN_SYNDROME_DIR / "schema.json"
DEFINITIONS_DIR = OPEN_SYNDROME_DIR / "definitions"
DEFINITIONS_DIR.mkdir(parents=True, exist_ok=True)


def download_schema():
    schema_response = requests.get(
        "https://raw.githubusercontent.com/OpenSyndrome/schema/refs/heads/main/schemas/"
        f"{OPEN_SYNDROME_VERSION}/schema.json"
    )
    SCHEMA_DIR.write_text(json.dumps(schema_response.json()))
    return SCHEMA_DIR


def download_definitions(url=None, current_path=None):
    """Download definitions from GitHub repos."""
    if url is None:
        url = "https://api.github.com/repos/OpenSyndrome/definitions/contents/definitions/v1?ref=main"
    response = requests.get(url)
    response.raise_for_status()

    for item in response.json():
        if item["type"] == "file":
            response_file = requests.get(item["download_url"])
            definition_filepath = current_path / item["name"]
            definition_filepath.write_bytes(response_file.content)

        elif item["type"] == "dir":
            print(item["path"])
            local_dir = DEFINITIONS_DIR / Path(item["path"]).parts[-1]
            local_dir.mkdir(parents=True, exist_ok=True)
            download_definitions(item["url"], local_dir)


def get_definition_dir():
    if not any(DEFINITIONS_DIR.iterdir()):
        download_definitions()
    return DEFINITIONS_DIR


def get_schema_filepath():
    if not SCHEMA_DIR.exists():
        download_schema()
    return SCHEMA_DIR


def load_definition(definition_filename):
    letter_dir = definition_filename[0].lower()
    return json.loads(
        (get_definition_dir() / f"{letter_dir}/{definition_filename}.json").read_text()
    )


def overlap_definitions(definitions):
    if len(definitions) < 2:
        return
    definition_codes = {}
    for definition in definitions:
        json_definition = load_definition(definition)
        inclusion_criteria = json_definition.get("inclusion_criteria", [])
        if inclusion_criteria:
            criterion = inclusion_criteria[0]
            definition_codes[definition] = set()
            for value in criterion["values"]:
                if value.get("code"):
                    definition_codes[definition].add(
                        value["code"]["code"].replace(".", "").upper()
                    )
    in_common_codes = reduce(lambda acc, cond: acc & cond, definition_codes.values())
    return in_common_codes


def filter_cases_per_definitions(df, mapping, definitions):
    _df_filtered = df
    for definition in definitions:
        _df_filtered = filter_cases(_df_filtered, mapping, definition)
    return _df_filtered


def find_cases_from(term):  # TODO rename
    definitions = []
    for file_ in get_definition_dir().glob("**/*.json"):
        if term.lower() in file_.name.lower():
            definitions.append(file_.name.replace(".json", ""))
    return definitions


def filter_cases(df, mapping, definition_filename):
    """Filter records in a DataFrame based on a definition file.

    The logical operator is not taking into account because the rows are seeing as independent.
    A new column is added to the DataFrame with the name of the definition file.
    """
    definition = load_definition(definition_filename)
    inclusion_criteria = definition.get("inclusion_criteria", [])
    # TODO add exclusion criteria
    if inclusion_criteria:
        criterion = inclusion_criteria[0]
        wanted_codes_per_system = {}
        for value in criterion["values"]:
            if value.get("code"):
                target_code = value["code"]["code"]
                system = value["code"]["system"]
                wanted_codes_per_system.setdefault(system, [])
                wanted_codes_per_system[system].append(target_code)

        all_filters = []
        print("Filtering records based on the following codes:")
        for system, codes in wanted_codes_per_system.items():
            for column in mapping:
                if column["system"] == system:
                    print(column["system"], codes)
                    regex_code = []
                    for code in codes:
                        if "%" in code:
                            # handle LIKE wildcard
                            regex_code.append(code.replace("%", ".*"))
                        else:
                            regex_code.append(f"{code}$")  # mark the end of the string
                    regex_code = "|".join(regex_code)
                    all_filters.append(pl.col(column["code"]).str.contains(regex_code))

        # FIXME show error if no filters are created all_filters == []
        combined_filter = reduce(lambda acc, cond: acc | cond, all_filters)
        return df.with_columns(combined_filter.alias(definition_filename))
    return df



from pathlib import Path
from unittest import mock
from unittest.mock import Mock, call

from opensyndrome.filtering import (
    download_definitions,
    filter_cases,
    find_cases_from,
    overlap_definitions,
    get_definition_dir,
    download_schema,
    get_schema_filepath,
)
import polars as pl


@mock.patch(
    "opensyndrome.filtering.get_definition_dir",
    return_value=Path("tests/definitions/v1"),
)
class TestFilterRecordsBasedOnDefinition:
    def test_filter_records_when_the_same_column_is_targeted(
        self, mock_definitions_dir
    ):
        df = pl.DataFrame(
            {
                "week": [1, 1, 1, 2, 2, 2, 1],
                "year": [
                    2025,
                    2025,
                    2025,
                    2024,
                    2024,
                    2024,
                    2024,
                ],
                "my_icd_code": [
                    "J109",
                    "J101",
                    "A929",
                    "A922",
                    "A929",
                    "U071",
                    "A929",
                ],
            }
        )
        mapping = [
            {"system": "ICD-10", "code": "my_icd_code"},
        ]
        assert df.shape == (7, 3)
        df = filter_cases(df, mapping, "arbovirosis_paraguay_sd")

        assert df.shape == (7, 4)
        assert "arbovirosis_paraguay_sd" in df.columns
        assert df["arbovirosis_paraguay_sd"].sum() == 3

    def test_filter_records_when_multiple_columns_are_targeted(
        self, mock_definitions_dir
    ):
        df = pl.DataFrame(
            {
                "week": [1, 1, 1, 2, 2, 2, 1],
                "year": [
                    2025,
                    2025,
                    2025,
                    2024,
                    2024,
                    2024,
                    2024,
                ],
                "my_icd_code": [
                    None,
                    "J101",
                    "A929",
                    "A922",
                    "A929",
                    "U071",
                    "A929",
                ],
                "ciap": [
                    "A77",
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                ],
            }
        )
        mapping = [
            {"system": "CID 10", "code": "my_icd_code"},
            {"system": "CID", "code": "my_icd_code"},
            {"system": "CIAP", "code": "ciap"},
        ]
        assert df.shape == (7, 4)
        df = filter_cases(df, mapping, "arbovirosis_aesop_brazil_sd")

        assert df.shape == (7, 5)
        assert "arbovirosis_aesop_brazil_sd" in df.columns
        assert df["arbovirosis_aesop_brazil_sd"].sum() == 4

    def test_filter_records_using_like_condition(self, mock_definitions_dir):
        df = pl.DataFrame(
            {
                "week": [1, 1, 1, 2, 2, 2, 1],
                "year": [
                    2025,
                    2025,
                    2025,
                    2024,
                    2024,
                    2024,
                    2024,
                ],
                "my_icd_code": [
                    None,
                    "J101",
                    "A972",
                    "A922",
                    "A929",
                    "U071",
                    "A979",
                ],
                "ciap": [
                    "A77",
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                ],
            }
        )
        mapping = [
            {"system": "CID 10", "code": "my_icd_code"},
        ]
        assert df.shape == (7, 4)
        df = filter_cases(df, mapping, "arbovirosis_aesop_brazil_sd_with_like")

        assert df.shape == (7, 5)
        assert "arbovirosis_aesop_brazil_sd_with_like" in df.columns
        assert df["arbovirosis_aesop_brazil_sd_with_like"].sum() == 2

    def test_filter_records_when_column_already_exists(self, mock_definitions_dir):
        df = pl.DataFrame(
            {
                "week": [1, 1, 1, 2, 2, 2, 1],
                "year": [
                    2025,
                    2025,
                    2025,
                    2024,
                    2024,
                    2024,
                    2024,
                ],
                "my_icd_code": [
                    "J109",
                    "J101",
                    "A929",
                    "A922",
                    "A929",
                    "U071",
                    "A929",
                ],
            }
        )
        mapping = [
            {"system": "ICD-10", "code": "my_icd_code"},
        ]
        assert df.shape == (7, 3)
        df_filtered = filter_cases(df, mapping, "arbovirosis_paraguay_sd")

        assert df_filtered.shape == (7, 4)
        assert "arbovirosis_paraguay_sd" in df_filtered.columns

        df_filtered_again = filter_cases(
            df_filtered, mapping, "arbovirosis_paraguay_sd"
        )

        assert df_filtered_again.shape == (7, 4)
        assert "arbovirosis_paraguay_sd" in df_filtered_again.columns


@mock.patch(
    "opensyndrome.filtering.get_definition_dir",
    return_value=Path("tests/definitions/v1"),
)
class TestCalculateOverlapAmongDefinitions:
    def test_calculate_overlap(self, mock_definitions_dir):
        definitions = find_cases_from("arbovirosis")
        assert len(definitions) == 3
        assert overlap_definitions(definitions) == {"A929"}

    def test_return_none_if_definitions_are_not_greater_than_two(
        self, mock_definitions_dir
    ):
        definitions = find_cases_from("covid")
        assert len(definitions) == 1
        assert overlap_definitions(definitions) is None


@mock.patch("opensyndrome.filtering.DEFINITIONS_DIR")
@mock.patch("opensyndrome.filtering.download_definitions")
class TestGetDefinitionsDir:
    def test_return_definitions_dir_if_not_empty(self, mock_download, mock_dir):
        mock_dir.iterdir.return_value = ["schema.json", "v1/"]

        get_definition_dir()

        assert mock_dir.iterdir.called
        assert mock_download.called is False

    def test_download_definitions_from_repo_if_dir_is_empty(
        self, mock_download, mock_dir
    ):
        mock_dir.iterdir.return_value = []

        get_definition_dir()

        assert mock_dir.iterdir.called
        assert mock_download.called is True


class TestDownloadSchema:
    @mock.patch("opensyndrome.filtering.SCHEMA_DIR")
    @mock.patch("opensyndrome.filtering.requests")
    def test_download_schema_from_github_repo(self, mock_requests, mock_dir):
        response = Mock()
        response.json.return_value = {"version": "1.0.0"}  # fake schema
        mock_requests.get.return_value = response

        download_schema()

        assert mock_requests.get.called
        assert mock_dir.mock_calls == [call.write_text('{"version": "1.0.0"}')]


@mock.patch("opensyndrome.filtering.SCHEMA_DIR")
@mock.patch("opensyndrome.filtering.download_schema")
class TestGetSchemaFilepath:
    def test_return_schema_filepath_if_exists(self, mock_download, mock_dir):
        mock_dir.exists.return_value = True

        get_schema_filepath()

        assert mock_dir.exists.called
        assert mock_download.called is False

    def test_download_schema_from_repo_if_dir_does_not_exist(
        self, mock_download, mock_dir
    ):
        mock_dir.exists.return_value = False

        get_schema_filepath()

        assert mock_dir.exists.called
        assert mock_download.called is True


@mock.patch("opensyndrome.filtering.requests")
@mock.patch("opensyndrome.filtering.DEFINITIONS_DIR")
class TestDownloadDefinitions:
    def test_download_definitions_recursively(self, mock_path, mock_requests):
        mock_response_v1 = Mock()
        mock_response_v1.json.return_value = [
            # some keys were omitted from the response due to its size
            {
                "name": "a",
                "path": "definitions/v1/a",
                "url": "https://api.github.com/repos/OpenSyndrome/definitions/contents/definitions/v1/a?ref=main",
                "type": "dir",
            }
        ]
        mock_response_a = Mock()
        mock_response_a.json.return_value = [
            {
                "name": "acuteflaccidparalysis_kenya.json",
                "path": "definitions/v1/a/acuteflaccidparalysis_kenya.json",
                "url": "https://api.github.com/repos/OpenSyndrome/definitions/contents/definitions/v1/a/acuteflaccidparalysis_kenya.json?ref=main",
                "download_url": "https://raw.githubusercontent.com/OpenSyndrome/definitions/main/definitions/v1/a/acuteflaccidparalysis_kenya.json",
                "type": "file",
            }
        ]
        mock_response_file = Mock()
        mock_response_file.content = b'{"key": "value"}'

        mock_requests.get.side_effect = [
            mock_response_v1,
            mock_response_a,
            mock_response_file,
        ]

        mock_path.return_value.write_bytes.return_value = None
        mock_path.return_value.mkdir.return_value = None

        download_definitions()

        assert mock_requests.get.call_count == 3
        calls = [
            call(
                "https://api.github.com/repos/OpenSyndrome/definitions/contents/definitions/v1?ref=main"
            ),
            call(
                "https://api.github.com/repos/OpenSyndrome/definitions/contents/definitions/v1/a?ref=main"
            ),
            call(
                "https://raw.githubusercontent.com/OpenSyndrome/definitions/main/definitions/v1/a/acuteflaccidparalysis_kenya.json"
            ),
        ]
        mock_requests.get.assert_has_calls(calls, any_order=True)

        assert (
            call.__truediv__().mkdir(parents=True, exist_ok=True)
            in mock_path.mock_calls
        )
        assert (
            call.__truediv__().__truediv__().write_bytes(b'{"key": "value"}')
            in mock_path.mock_calls
        )



import marimo

__generated_with = "0.14.5"
app = marimo.App(
    width="medium",
    app_title="Open Syndrome Definition - Data Browser",
)


@app.cell
def _():
    import marimo as mo
    import json
    from pathlib import Path

    from flatten_json import flatten
    import polars as pl
    import plotly.graph_objects as go
    from dotenv import load_dotenv

    from opensyndrome.filtering import (
        filter_cases_per_definitions,
        load_definition,
        overlap_definitions,
        get_definition_dir,
    )

    return (
        Path,
        filter_cases_per_definitions,
        flatten,
        go,
        json,
        load_definition,
        load_dotenv,
        mo,
        overlap_definitions,
        pl,
        get_definition_dir,
    )


@app.cell(hide_code=True)
def _(load_dotenv):
    result = load_dotenv()  # noqa
    return


@app.cell
def _(go, pl):
    def plot_cases(_df_filtered, definitions, date_column="date"):
        _definitions_columns_sum = [
            pl.col(definition).sum().alias(definition) for definition in definitions
        ]
        _agg_df = (
            _df_filtered.group_by(date_column)
            .agg(_definitions_columns_sum)
            .sort(date_column)
        )
        _fig = go.Figure()
        for definition in definitions:
            _fig.add_trace(
                go.Scatter(
                    x=_agg_df[date_column],
                    y=_agg_df[definition],
                    mode="lines+markers",
                    name=definition,
                )
            )
        return _fig

    return (plot_cases,)


@app.cell
def _(go, pl):
    def groupped_bar(_df_filtered, definitions, group_by_column="code", top_n=10):
        _agg_df = (
            _df_filtered.group_by(group_by_column)
            .agg([pl.col(definition).sum() for definition in definitions])
            .sort(group_by_column)
        )

        _agg_df = _agg_df.with_columns(
            (sum([pl.col(definition) for definition in definitions])).alias("total")
        )
        _agg_df = _agg_df.sort("total", descending=True).head(
            top_n
        )  # .sort(group_by_column)
        _fig = go.Figure(
            data=[
                go.Bar(
                    name=definition, x=_agg_df[group_by_column], y=_agg_df[definition]
                )
                for definition in definitions
            ]
        )
        _fig.update_layout(barmode="group")
        return _fig

    return (groupped_bar,)


@app.cell
def _(mo):
    mo.md(r"""# Open Syndrome Definition 👩🏽‍🔬""")
    return


@app.cell
def _(mo):
    mo.callout(
        "This is a prototype of how to filter your data using definitions from the Open Syndrome Initiative.",
        kind="neutral",
    )
    return


@app.cell
def _(json, mo):
    empty_mapping = [
        {"system": "CID 10", "code": "icd_code"},
        {"system": "ICD-10", "code": "icd_code"},
        {"system": "CID", "code": "icd_code"},
    ]
    mapping_editor = mo.ui.code_editor(
        value=json.dumps(empty_mapping, indent=2), language="javascript"
    )  # json

    mapping_form = (
        mo.md(
            """
        {date_column}

        {code_column}

        {mapping}
    """
        )
        .batch(
            date_column=mo.ui.text(label="Datetime column", placeholder="recording_ts"),
            code_column=mo.ui.text(
                label="Diagnosis code column", placeholder="icd_code"
            ),
            mapping=mapping_editor,
        )
        .form(bordered=True)
    )
    return (mapping_form,)


@app.cell
def _(mo):
    sample_file = mo.ui.file(
        kind="area", filetypes=[".csv"]
    )  # TODO max_size: int = 100000000 - the maximum file size is 100MB
    sample_file  # return is (name, contents)
    return (sample_file,)


@app.cell
def _(pl):
    def load_df_selected(sample_file):
        if not sample_file.value:
            return
        return pl.read_csv(sample_file.contents())

    return (load_df_selected,)


@app.cell
def _(load_df_selected, sample_file):
    df_selected = load_df_selected(sample_file)
    return (df_selected,)


@app.cell
def _(df_selected, mapping_form, mo):
    mo.stop(df_selected is None)

    mo.vstack(
        [
            mo.md("### **Data**"),
            mo.ui.dataframe(df_selected),
            mo.accordion(
                {
                    "Mapping configuration": mapping_form,
                }
            ),
        ]
    )
    return


@app.cell
def _(json, mapping_form, mo):
    mo.stop(mapping_form.value is None)

    date_column = mapping_form.value["date_column"]
    code_column = mapping_form.value["code_column"]
    mapping = json.loads(mapping_form.value["mapping"])
    return code_column, date_column, mapping


@app.cell
def _(get_definition_dir):
    definition_options = {
        filepath.name.replace(".json", ""): filepath
        for filepath in get_definition_dir().glob(
            "**/*.json"
        )  # TODO fetch from a remote URL
    }
    return (definition_options,)


@app.cell
def _(definition_options, mo):
    definitions_dropdown = mo.ui.multiselect(
        label="Select Syndromic Indicators", options=sorted(definition_options.keys())
    )
    return (definitions_dropdown,)


@app.cell
def _(flatten, go, load_definition, mo, pl):
    def plot_flatted_json(df):
        cols = df.columns
        _fig = go.Figure(
            data=[
                go.Table(
                    header=dict(values=cols),
                    cells=dict(
                        values=[df[col] for col in cols],
                        fill_color="lavender",
                        align="left",
                    ),
                )
            ]
        )
        _fig.show()
        return _fig

    def flat_df(json_data, name):
        flat = flatten(json_data, separator=".")
        return pl.DataFrame(
            {
                "definition": name or json_data.get("title"),
                "field": flat.keys(),
                "value": flat.values(),
            },
            strict=False,
        )

    def flat_definitions(definitions):
        flatted = [
            flat_df(
                {
                    "inclusion_criteria": load_definition(definition_name)[
                        "inclusion_criteria"
                    ]
                },
                definition_name,
            )
            for definition_name in definitions
        ]
        table = pl.concat(flatted, how="vertical_relaxed").pivot(
            index=["definition"], on="field"
        )
        table = table.transpose(include_header=True, header_name="fields")
        table.rename(dict(zip(table.columns[1:], table.row(0)[1:]))).slice(1)
        return plot_flatted_json(table)  # FIXME

    def plot_comparison_between_definitions(definitions, overlapped):
        # return mo.ui.plotly(flat_definitions(definitions))
        return mo.md(",".join(overlapped) if overlapped else "-")

    return (plot_comparison_between_definitions,)


@app.cell
def _(definitions_dropdown, mo):
    mo.hstack([mo.md("**::lucide:filter:: Filters:**"), definitions_dropdown])
    return


@app.cell
def _(
    definitions_dropdown,
    df_selected,
    filter_cases_per_definitions,
    mapping,
    mo,
    overlap_definitions,
):
    mo.stop(
        df_selected is None
        or df_selected.is_empty()
        or definitions_dropdown.value is None
        or mapping is None
    )

    definitions = None
    df_filtered = None
    overlapped = None

    if not df_selected.is_empty():
        definitions = definitions_dropdown.value
        df_filtered = filter_cases_per_definitions(df_selected, mapping, definitions)
        overlapped = overlap_definitions(definitions)
    return definitions, df_filtered, overlapped


@app.cell
def _(definitions, df_filtered, df_selected, mo, overlapped):
    mo.stop(definitions is None or df_filtered is None)

    _cards = [
        mo.stat(
            label="Rows & columns",
            value=df_selected.shape[0],
        ),
        mo.stat(
            label="Columns",
            value=df_selected.shape[1],
        ),
        mo.stat(
            label="Syndromic Indicators",
            value=len(definitions),
            caption=", ".join([definition for definition in definitions]),
            bordered=True,
        ),
        mo.stat(
            label="Number of common criteria",
            value=len(overlapped) if overlapped else "-",
        ),
    ]

    _title = "## Data with Open Syndrome Definitions"

    mo.vstack(
        [
            mo.md(_title),
            mo.hstack(_cards, widths="equal", align="center"),
        ]
    )
    return


@app.cell
def _(
    definitions,
    load_definition,
    mo,
    overlapped,
    plot_comparison_between_definitions,
):
    mo.stop(definitions is None)

    mo.vstack(
        [
            mo.md("### Definitions details"),
            mo.md(
                "This section shows the definitions used to filter the data. You can use them to understand how the data was filtered and what criteria were applied. 🔎"
            ),
            mo.ui.tabs(
                {
                    "In common": plot_comparison_between_definitions(
                        definitions, overlapped
                    ),
                    "JSONs": mo.accordion(
                        {
                            definition: mo.json(load_definition(definition))
                            for definition in definitions
                        }
                    ),
                },
            ),
        ]
    )
    return


@app.cell
def _(mo):
    top_n = mo.ui.number(start=1, stop=10, label="Number of top codes", value=3, step=1)
    return (top_n,)


@app.cell
def _(
    code_column,
    date_column,
    definitions,
    df_filtered,
    df_selected,
    groupped_bar,
    mo,
    plot_cases,
    top_n,
):
    mo.stop(definitions is None or df_selected is None)

    mo.vstack(
        [
            mo.md("## Time series"),
            plot_cases(df_filtered, definitions, date_column=date_column),
            mo.md("## Codes comparison per syndromic indicator"),
            top_n.left(),
            groupped_bar(
                df_filtered,
                definitions,
                top_n=top_n.value or 3,
                group_by_column=code_column,
            ),
        ]
    )
    return


if __name__ == "__main__":
    app.run()
